// Copyright 2024 Pierre Lairez
// SPDX-License-Identifier: GPL-3.0-or-later

use algpath::{
    cli::Arithmetic,
    flint::Fmpq,
    prog::{ExactProg, FlintProg, ProgramTrait, RecklessProg},
    projective::{ParametricSystem, ProjectivePoint},
    traits::{
        system,
        track::{
            find_moore_box::{ArbFindMooreBox, ArbRecklessFindMooreBox, FindMooreBox, RecklessFindMooreBox},
            ArbReckless5Track, ArbTrack, RecklessTrack, RecklessTrack5, Track,
        },
        vector::Vector,
    },
};
use clap::Parser;
use flint_bindings::flint_cleanup;
use nalgebra::Complex;
use num_traits::{One, Zero};
use serde_json::json;
use std::{
    ops::Deref,
    sync::{Arc, Mutex},
    thread,
    time::Instant,
};

fn auxi<
    Prog,
    TrackAlgo: Track,
    FindMb: FindMooreBox<Sys = TrackAlgo::Sys, C = TrackAlgo::C, VecC = TrackAlgo::VecC, MatC = TrackAlgo::MatC>,
>(
    exact_sys: Vec<ParametricSystem<ExactProg>>,
    fiber: Vec<Vec<Complex<Fmpq>>>,
    num_jobs: usize,
    before_compilation: Instant,
) where
    Prog: ProgramTrait<Coeffs = TrackAlgo::C>,
    for<'r> TrackAlgo::C: From<&'r Complex<Fmpq>>,
    TrackAlgo::C: Zero,
    TrackAlgo::VecC: From<Vec<TrackAlgo::C>> + Vector<T = TrackAlgo::C>,
    TrackAlgo::Sys: From<ParametricSystem<ExactProg>> + Send + Sync + 'static,
{
    let f_instructions = exact_sys[0].f[0].instructions.len();
    let df_instructions = exact_sys[0].df[0].instructions.len();
    // Build the systems in the arithmetic from the exact systems
    let systems: Arc<Vec<TrackAlgo::Sys>> = Arc::new(exact_sys.into_iter().map(|f| f.into()).collect());

    let card_fiber = fiber.len();
    let indices: Vec<_> = (0..card_fiber).collect();
    let fiber = Arc::new(fiber);
    let workq = Arc::new(Mutex::new(indices));

    let after_compilation = Instant::now();

    let mut handles = Vec::new();
    let allresults = Arc::new(Mutex::new(Vec::new()));
    for _ in 0..num_jobs {
        let workq = workq.clone();
        let allresults = allresults.clone();
        let fiber = fiber.clone();
        let systems = systems.clone();
        let handle = thread::spawn(move || {
            let mut results = Vec::new();
            loop {
                let mut workq = workq.lock().unwrap();
                let point_index = match workq.pop() {
                    Some(p) => p,
                    None => break,
                };
                let point = &fiber[point_index];
                let n = card_fiber - workq.len();
                eprintln!("Starting continuation {}/{}...", n, card_fiber);
                std::mem::drop(workq);

                let proj_point = ProjectivePoint {
                    coord: point.clone().iter().map(|c| c.into()).collect::<Vec<TrackAlgo::C>>().into(),
                    chart: 0,
                };

                let mut mbox = FindMb::find_moore_box(&systems[0], &vec![TrackAlgo::C::zero()].into(), proj_point).unwrap();

                for (i, system) in systems.iter().enumerate() {
                    eprintln!("Continuation {}/{}, segment {}/{}", n, card_fiber, i + 1, systems.len());

                    let tropt = TrackAlgo::track(system, &mut mbox);
                    match tropt {
                        Ok(tr) => {
                            mbox.params = TrackAlgo::VecC::from_value(1, TrackAlgo::C::zero());
                            eprintln!("{}", tr.len());
                            results.push((
                                point_index,
                                Some((
                                    tr.len(),
                                    tr.iter().map(|x| x.2).max().unwrap(), // Max precision (in bits) during the continuation
                                    tr.iter().map(|x| f64::from(u32::try_from(x.2).unwrap())).sum::<f64>(), // Sum of the precisions
                                    tr.iter().map(|x| f64::from(u32::try_from(x.2).unwrap()) * x.3).sum::<f64>()
                                        / tr.iter().map(|x| x.3).sum::<f64>(), // Weighted (by the stepsize) sum of the precisions
                                )),
                            ));
                        }
                        Err(e) => {
                            eprintln!("{}", e);
                            results.push((point_index, None));
                        }
                    }
                }
                // eprintln!("continuation {}/{} done !", n, card_fiber);
            }
            let mut allresults = allresults.lock().unwrap();
            allresults.append(&mut results);
        });
        handles.push(handle);
    }

    for handle in handles {
        handle.join().unwrap();
    }

    let elapsed = after_compilation.elapsed();
    let elapsed_with_overhead = before_compilation.elapsed();

    let mut allresults = allresults.lock().unwrap();
    allresults.reverse();

    allresults.sort_by_key(|t| t.0);

    // Compute some statistics
    let mut iter_list: Vec<f64> = Vec::new();
    let mut max_prec_list: Vec<f64> = Vec::new();
    let mut mean_prec_list: Vec<f64> = Vec::new();
    let mut weighted_mean_prec_list: Vec<f64> = Vec::new();
    let mut failures = 0usize;
    for (_, r) in allresults.deref() {
        match r {
            Some((iters, max_prec, mean_prec, weighted_mean_prec)) => {
                let iters_f64 = u32::try_from(*iters).unwrap().into();
                let max_prec_f64 = u32::try_from(*max_prec).unwrap().into();
                iter_list.push(iters_f64);
                max_prec_list.push(max_prec_f64);
                mean_prec_list.push(*mean_prec / iters_f64);
                weighted_mean_prec_list.push(*weighted_mean_prec);
            }
            None => {
                iter_list.push(f64::NAN);
                failures += 1;
            }
        }
    }

    let output = json!({
        "instructions": {
            "f": f_instructions,
            "df": df_instructions,
        },
        "steplist": iter_list,
        "failures": failures,
        "time": elapsed.as_secs_f32(),
        "overheadtime": elapsed_with_overhead.as_secs_f32(),
        "maxpreclist": max_prec_list,
        "meanpreclist": mean_prec_list,
        "weightedmeanpreclist": weighted_mean_prec_list
    });

    println!("{}", output["time"]);
}

fn main() {
    let conf = algpath::cli::Configuration::parse();
    if conf.markdown_help {
        clap_markdown::print_help_markdown::<algpath::cli::Configuration>();
        return;
    }

    use std::time::Instant;
    let before_compilation = Instant::now();

    let problem = algpath::cli::Problem::new_from_input(&conf).unwrap();
    let mut problems = if problem.path == vec![vec![Complex::<Fmpq>::zero(), Complex::<Fmpq>::one()]] {
        vec![problem.clone()]
    } else {
        problem.precompose_path().unwrap()
    };

    let parametric_system = problems
        .iter_mut()
        .map(|problem| problem.build_parametric_system(&conf).unwrap())
        .collect();

    let num_jobs: usize;
    if let Some(n) = conf.jobs {
        num_jobs = n;
    } else if let Ok(n) = thread::available_parallelism() {
        num_jobs = n.into();
    } else {
        num_jobs = 1;
    }

    match conf.arithmetic {
        Arithmetic::Reckless => {
            auxi::<RecklessProg, RecklessTrack5, RecklessFindMooreBox>(
                parametric_system,
                problem.fiber,
                num_jobs,
                before_compilation,
            );
        }
        Arithmetic::Arb => {
            auxi::<FlintProg, ArbTrack<5>, ArbFindMooreBox>(parametric_system, problem.fiber, num_jobs, before_compilation);
        }
        Arithmetic::ArbReckless => {
            auxi::<FlintProg, ArbReckless5Track, ArbRecklessFindMooreBox>(
                parametric_system,
                problem.fiber,
                num_jobs,
                before_compilation,
            );
        }
    }

    unsafe {
        flint_cleanup();
    }
}
