#![allow(dead_code)]
// Copyright 2024 Pierre Lairez
// SPDX-License-Identifier: GPL-3.0-or-later

use super::circuit;
use crate::circuit::Base;
use crate::flint::Acb;
use crate::flint::Fmpq;
use crate::precision_ctx::get_prec;
use crate::reckless::with_rounding_up;
use crate::reckless::Cif;
use crate::reckless_extra::RecklessTm5;
use crate::reckless_extra::UnsafeCif;
use crate::reckless_extra::UnsafeRecklessTm5;
use crate::taylor_model::tm::RecklessTm;
use crate::traits::num::RefNeg;
use crate::traits::vector::Vector;
use anymap::AnyMap;
use circuit::Circuit;
use circuit::Node;
use circuit::Visitor;
use nalgebra::Complex;
use num_traits::Zero;
use std::cell::RefCell;
use std::collections::HashMap;
use std::collections::HashSet;
use std::iter::repeat;
use std::ops::AddAssign;
use std::ops::MulAssign;
use std::sync::Arc;
use std::sync::RwLock;

#[derive(Debug, Clone, Copy)]
pub enum Instruction {
    Move(usize, usize),
    Add(usize, usize),
    Mul(usize, usize),
    AddScalar(usize, usize),
    MulScalar(usize, usize),
    Neg(usize),
    Scalar(usize, usize),
    Zero(usize),
    Nop,
}

use Instruction::*;

#[derive(Debug)]
pub struct Program<T> {
    pub scalars: RwLock<Arc<Vec<T>>>,
    pub instructions: Vec<Instruction>,
    pub vars: Vec<String>,
    pub outnodes: Vec<usize>,
    pub bufferlen: usize,
}

unsafe impl<T> Send for Program<T> {}
unsafe impl<T> Sync for Program<T> {}

pub struct _Program<T> {
    pub scalars: Vec<T>,
    pub instructions: Vec<Instruction>,
    pub vars: Vec<String>,
    pub outnodes: Vec<usize>,
    pub bufferlen: usize,
}

impl<T> Program<T> {
    pub fn get_scalars(&self) -> Arc<Vec<T>> {
        self.scalars.read().unwrap().clone()
    }

    pub fn set_scalars(&self, scalars: Vec<T>) {
        *self.scalars.write().unwrap() = Arc::new(scalars);
    }
}

thread_local! {
    static BUFFERS: RefCell<AnyMap> = RefCell::new(AnyMap::new());
}

fn resize_buffer<S: Clone + Zero + 'static>(size: usize) {
    BUFFERS.with(|map_ref: &RefCell<AnyMap>| {
        let mut map = map_ref.borrow_mut();
        if map.get::<Vec<S>>().is_none() {
            // dbg!("Initializing buffer");
            map.insert(repeat(S::zero()).take(size).collect::<Vec<S>>());
        } else {
            let buffer = map.get_mut::<Vec<S>>().unwrap();
            if buffer.len() < size {
                // dbg!(size);
                buffer.resize(size, S::zero());
            }
        }
    })
}

// impl<S, T> Set<&Vec<S>> for Program<T>
// where
//     for<'a> T: Set<&'a S>,
// {
//     fn set(&mut self, other: &Vec<S>) {
//         debug_assert!(other.len() >= self.scalars.len());
//         self.scalars.iter_mut().zip(other.iter()).for_each(|(t, s)| t.set(s));
//     }
// }

impl<T> Program<T> {
    // pub fn from_program<S: Into<T>>(prog: Program<S>) -> Program<T> {
    //     Program {
    //         scalars: prog.scalars.into_iter().map(|c| c.into()).collect(),
    //         instructions: prog.instructions,
    //         vars: prog.vars,
    //         outnodes: prog.outnodes,
    //         bufferlen: prog.bufferlen,
    //     }
    // }

    pub fn compile<'a, I: IntoIterator<Item = &'a Circuit<T>>>(circuits: I, vars: I) -> Self
    where
        T: 'a + Base + Zero,
    {
        // 1. Compute the sequence of all node to be computed. Gather the output
        // nodes (which are the elements of the inputs of this function) in
        // outnodeset.
        let mut ord = circuit::TopologicalOrderingVisitor {
            nodes: Default::default(),
        };

        let mut outnodeset = HashSet::new();
        let mut outnodelist = Vec::new();
        let mut memo = Default::default();
        for circ in circuits {
            ord.compute(circ, &mut memo);
            outnodeset.insert(circ);
            outnodelist.push(circ.clone())
        }

        // 2. Initialize the program.
        let mut prog = _Program {
            scalars: Default::default(),
            instructions: Default::default(),
            vars: Default::default(),
            outnodes: Default::default(),

            // Length of the buffer that we will need. We make so basic effort to
            // reuse free cells, but reordering the computation to minimize
            // bufferlen is a difficult combinatorial optimization problem.
            bufferlen: 0,
        };

        // 3. First pass over the nodes.

        // Associate to each intermediate circuit, the cell of prog.buffer in
        // which it will be computed.
        let mut indices = HashMap::<&Circuit<T>, usize>::default();

        // The input variables are allocated the firt indices
        for (i, v) in vars.into_iter().enumerate() {
            indices.insert(v, i);
            match v.get_var_name() {
                None => panic!(),
                Some(s) => prog.vars.push(s.to_string()),
            }
        }
        prog.bufferlen = prog.vars.len();

        // Associate to each intermediate circuit the number of times it is used
        // afterwards.
        let mut nb_read = HashMap::<Circuit<T>, usize>::default();

        // iterate over the intermediate circuit, in topological order: every
        // argument of an operations are treated before the operation.
        for circ in &ord.nodes {
            nb_read.insert(circ.clone(), if outnodeset.contains(circ) { 1 } else { 0 });
            match circ.node.as_ref() {
                Node::Add(x, y) | Node::Mul(x, y) => {
                    // The first argument cannot be a scalar. This is enforced by
                    // Node::simplify.
                    assert!(x.get_scalar().is_none());
                    *nb_read.get_mut(x).unwrap() += 1;
                    *nb_read.get_mut(y).unwrap() += 1;
                }
                Node::Neg(x) => {
                    // Enforced by Node::simplify
                    assert!(x.get_scalar().is_none());
                    *nb_read.get_mut(x).unwrap() += 1;
                }
                Node::Scalar(_) | Node::Zero => (),
                Node::Var(_) => {
                    assert!(indices.contains_key(circ));
                }
            };
        }

        // 4. Second pass

        let mut scalars_indices = HashMap::<T, usize>::default();
        let mut emitscalar = |c: &T| {
            if let Some(i) = scalars_indices.get(c) {
                *i
            } else {
                let i = prog.scalars.len();
                scalars_indices.insert(c.clone(), i);
                prog.scalars.push(c.clone());
                i
            }
        };

        // Contains all the free cells
        let mut freelist = Vec::<usize>::new();
        // let mut forgotten_nodes = HashSet::new();

        for circ in &ord.nodes {
            //dbg!(circ);

            // Find the index at which the result of the computation will be
            // stored.
            let mut getidx = || {
                let idx;
                if let Some(i) = freelist.pop() {
                    idx = i;
                } else {
                    idx = prog.bufferlen;
                    prog.bufferlen += 1;
                }
                indices.insert(circ, idx);
                idx
            };

            match circ.node.as_ref() {
                Node::Add(x, y) => {
                    let idx = getidx();
                    if let Node::Scalar(s) = y.node.as_ref() {
                        if idx != indices[x] {
                            prog.instructions.push(Move(idx, indices[x]));
                        }
                        prog.instructions.push(AddScalar(idx, emitscalar(s)));
                    } else {
                        if idx == indices[x] {
                            prog.instructions.push(Add(idx, indices[y]));
                        } else {
                            if idx != indices[y] {
                                prog.instructions.push(Move(idx, indices[y]));
                            }
                            prog.instructions.push(Add(idx, indices[x]));
                        }
                    }
                }
                Node::Mul(x, y) => {
                    let idx = getidx();
                    if let Node::Scalar(s) = y.node.as_ref() {
                        if idx != indices[x] {
                            prog.instructions.push(Move(idx, indices[x]));
                        }
                        prog.instructions.push(MulScalar(idx, emitscalar(s)));
                    } else {
                        if idx == indices[x] {
                            prog.instructions.push(Mul(idx, indices[y]));
                        } else if idx == indices[y] {
                            prog.instructions.push(Mul(idx, indices[x]));
                        } else {
                            prog.instructions.push(Move(idx, indices[x]));
                            prog.instructions.push(Mul(idx, indices[y]));
                        }
                    }
                }

                Node::Neg(x) => {
                    let idx = getidx();
                    if idx != indices[x] {
                        prog.instructions.push(Move(idx, indices[x]));
                    }
                    prog.instructions.push(Neg(idx));
                }
                Node::Var(_) => (),
                Node::Scalar(s) => {
                    if outnodeset.contains(circ) {
                        let idx = getidx();
                        prog.instructions.push(Scalar(idx, emitscalar(s)));
                    }
                }
                Node::Zero => {
                    let idx = getidx();
                    prog.instructions.push(Zero(idx));
                }
            }

            // Decrement the "nbread" counters and update freelist accordingly.
            match circ.node.as_ref() {
                Node::Add(x, y) | Node::Mul(x, y) => {
                    let rx = nb_read.get_mut(x).unwrap();
                    *rx -= 1;
                    if *rx == 0 {
                        if let Some(i) = indices.get(x) {
                            freelist.push(*i);
                            indices.remove(x);
                            //dbg!(&indices);
                        }
                    }
                    let ry = nb_read.get_mut(y).unwrap();
                    *ry -= 1;
                    if *ry == 0 {
                        if let Some(i) = indices.get(y) {
                            freelist.push(*i);
                            indices.remove(y);
                            //dbg!(&indices);
                        }
                    }
                }
                Node::Neg(x) => {
                    let rx = nb_read.get_mut(x).unwrap();
                    *rx -= 1;
                    if *rx == 0 {
                        if let Some(i) = indices.get(x) {
                            freelist.push(*i);
                            indices.remove(x);
                            //dbg!(&indices);
                        }
                    }
                }
                Node::Var(_) | Node::Scalar(_) | Node::Zero => (),
            }
        }

        prog.outnodes.reserve_exact(outnodelist.len());
        for circ in &outnodelist {
            prog.outnodes.push(indices[circ]);
        }

        return Program {
            scalars: RwLock::new(Arc::new(prog.scalars)),
            instructions: prog.instructions,
            vars: prog.vars,
            outnodes: prog.outnodes,
            bufferlen: prog.bufferlen,
        };
    }
}

pub trait ProgramTrait {
    type Coeffs;

    fn bufferlen(&self) -> usize;
    fn outnodes(&self) -> &Vec<usize>;
}

impl<T> ProgramTrait for Program<T> {
    type Coeffs = T;

    fn bufferlen(&self) -> usize {
        self.bufferlen
    }

    fn outnodes(&self) -> &Vec<usize> {
        &self.outnodes
    }
}

pub trait Eval<In, Out> {
    fn eval(&self, input: In) -> Out;
}

impl<Prog, S> Eval<(), Vec<S>> for Prog
where
    S: Clone + 'static,
    Self: ProgramTrait,
    for<'a> Self: Eval<&'a mut Vec<S>, ()>,
{
    fn eval(&self, _: ()) -> Vec<S> {
        let mut result = Vec::default();

        BUFFERS.with(|map_ref| {
            let mut map = map_ref.borrow_mut();
            let buffer: &mut Vec<S> = match map.get_mut::<Vec<S>>() {
                None => panic!("Buffer was not allocated"),
                Some(v) => v,
            };
            assert!(
                buffer.len() >= self.bufferlen(),
                "Buffer should be initialized with proper size"
            );

            self.eval(buffer);

            result.reserve_exact(self.outnodes().len());
            for i in self.outnodes() {
                result.push(buffer[*i].clone());
            }
        });

        result
    }
}

impl<T, S> Eval<&mut Vec<S>, ()> for Program<T>
where
    S: Clone
        + Zero
        + RefNeg
        + for<'a> AddAssign<&'a S>
        + for<'a> MulAssign<&'a S>
        + for<'a> AddAssign<&'a T>
        + for<'a> MulAssign<&'a T>
        + 'static,
    T: Clone + Into<S>,
{
    fn eval(&self, buffer: &mut Vec<S>) {
        let tmp_scalars = self.get_scalars();
        for instr in &self.instructions {
            match instr {
                Move(dest, x) => unsafe { *buffer.get_unchecked_mut(*dest) = buffer.get_unchecked(*x).clone() },
                Add(dest, x) => unsafe {
                    let a = buffer[*x].clone();
                    *buffer.get_unchecked_mut(*dest) += &a;
                },
                Mul(dest, x) => unsafe {
                    let a = buffer[*x].clone();
                    *buffer.get_unchecked_mut(*dest) *= &a;
                },
                Neg(dest) => unsafe {
                    buffer.get_unchecked_mut(*dest).ref_neg();
                },
                AddScalar(dest, i) => unsafe {
                    *buffer.get_unchecked_mut(*dest) += tmp_scalars.get_unchecked(*i);
                },
                MulScalar(dest, i) => unsafe {
                    *buffer.get_unchecked_mut(*dest) *= tmp_scalars.get_unchecked(*i);
                },
                Scalar(dest, i) => unsafe { *buffer.get_unchecked_mut(*dest) = (tmp_scalars.get_unchecked(*i).clone()).into() },
                Zero(dest) => unsafe { *buffer.get_unchecked_mut(*dest) = S::zero() },
                Nop => (),
            }
        }
    }
}

impl<S, T> Eval<&HashMap<String, S>, Vec<S>> for Program<T>
where
    for<'r> Program<T>: Eval<(), Vec<S>>,
    S: Clone + Zero + 'static,
{
    fn eval(&self, input: &HashMap<String, S>) -> Vec<S> {
        resize_buffer::<S>(self.bufferlen);
        BUFFERS.with(|map_ref: &RefCell<AnyMap>| {
            let mut map = map_ref.borrow_mut();
            let buffer: &mut Vec<S> = map.get_mut::<Vec<S>>().unwrap();

            // Setting the buffer
            for (i, v) in self.vars.iter().enumerate() {
                buffer[i] = input[v].clone();
            }
        });
        self.eval(())
    }
}

impl<Prog, S> Eval<&Vec<S>, Vec<S>> for Prog
where
    Self: ProgramTrait,
    Self: Eval<(), Vec<S>>,
    S: Clone + Zero + 'static,
{
    fn eval(&self, input: &Vec<S>) -> Vec<S> {
        resize_buffer::<S>(self.bufferlen());
        BUFFERS.with(|map_ref: &RefCell<AnyMap>| {
            let mut map = map_ref.borrow_mut();
            let buffer: &mut Vec<S> = map.get_mut::<Vec<S>>().unwrap();

            // Setting the buffer
            for i in 0..self.bufferlen().min(input.len()) {
                buffer[i] = input[i].clone();
            }
        });
        self.eval(())
    }
}

impl<VecS, Prog> Eval<(&VecS, &VecS), Vec<VecS::T>> for Prog
where
    Prog: Eval<(), Vec<VecS::T>> + ProgramTrait,
    VecS::T: Clone + Zero + 'static,
    VecS: Vector,
{
    fn eval(&self, (params, pt): (&VecS, &VecS)) -> Vec<VecS::T> {
        resize_buffer::<VecS::T>(self.bufferlen());
        BUFFERS.with(|map_ref: &RefCell<AnyMap>| {
            let mut map = map_ref.borrow_mut();
            let buffer: &mut Vec<VecS::T> = map.get_mut::<Vec<VecS::T>>().unwrap();

            // Setting the buffer
            for i in 0..params.len().min(self.bufferlen()) {
                buffer[i] = params[i].clone()
            }
            for j in 0..pt.len().min(self.bufferlen() - params.len()) {
                buffer[j + params.len()] = pt[j].clone();
            }
        });
        self.eval(())
    }
}

pub fn eval_map<F, T, Prog>(prog: &Prog, input_map: F) -> Vec<T>
where
    Prog: Eval<(), Vec<T>> + ProgramTrait,
    F: Fn(usize) -> Option<T>,
    T: Clone + Zero + 'static,
{
    resize_buffer::<T>(prog.bufferlen());
    BUFFERS.with(|map_ref: &RefCell<AnyMap>| {
        let mut map = map_ref.borrow_mut();
        let buffer: &mut Vec<T> = map.get_mut::<Vec<T>>().unwrap();

        // Setting the buffer

        for i in 0..prog.bufferlen() {
            if let Some(x) = input_map(i) {
                buffer[i] = x;
            } else {
                break;
            }
        }
    });
    prog.eval(())
}

pub struct RecklessProg(pub Program<UnsafeCif>);

impl ProgramTrait for RecklessProg {
    type Coeffs = Cif;

    fn bufferlen(&self) -> usize {
        self.0.bufferlen
    }

    fn outnodes(&self) -> &Vec<usize> {
        &self.0.outnodes
    }
}

impl Eval<&mut Vec<Cif>, ()> for RecklessProg {
    fn eval(&self, input: &mut Vec<Cif>) -> () {
        with_rounding_up(|| unsafe { self.0.eval(&mut *(input as *mut Vec<Cif> as *mut Vec<UnsafeCif>)) })
    }
}

impl Eval<&mut Vec<RecklessTm5>, ()> for RecklessProg {
    fn eval(&self, input: &mut Vec<RecklessTm5>) -> () {
        with_rounding_up(|| unsafe {
            self.0
                .eval(&mut *(input as *mut Vec<RecklessTm5> as *mut Vec<UnsafeRecklessTm5>))
        })
    }
}

impl<const N: usize> Eval<&mut Vec<RecklessTm<N>>, ()> for RecklessProg {
    fn eval(&self, input: &mut Vec<RecklessTm<N>>) -> () {
        unsafe { (&*(&self.0 as *const Program<UnsafeCif> as *const Program<Cif>)).eval(input) }
    }
}

pub type ExactProg = Program<Complex<Fmpq>>;

pub type FlintProg = Program<Acb>;

pub struct AcbProg {
    pub program: Program<Acb>,
    pub exact_coeffs: Vec<Complex<Fmpq>>,
    pub prec: RwLock<i64>,
}

impl AcbProg {
    pub fn update_prec(&self) {
        if *self.prec.read().unwrap() < get_prec() {
            // We require precision and scalars in write mode and update them simultaneously
            let mut prec_write = self.prec.write().unwrap();
            #[cfg(debug_assertions)]
            println!("Updating program coefficients ({} to {} bits of precision)...", prec_write, get_prec());
            let mut program_scalars_write = self.program.scalars.write().unwrap();
            *prec_write = get_prec();
            *program_scalars_write = Arc::new(self.exact_coeffs.iter().map(|scalar| Acb::from(scalar)).collect());
        }
    }
}

impl ProgramTrait for AcbProg {
    type Coeffs = Acb;

    fn bufferlen(&self) -> usize {
        self.program.bufferlen
    }

    fn outnodes(&self) -> &Vec<usize> {
        &self.program.outnodes
    }
}

impl<S> Eval<&mut Vec<S>, ()> for AcbProg
where
    for<'a> Program<Acb>: Eval<&'a mut Vec<S>, ()>,
{
    fn eval(&self, input: &mut Vec<S>) -> () {
        self.update_prec();
        self.program.eval(input)
    }
}

#[cfg(test)]
mod evaluation {
    use super::*;
    use crate::{
        flint::{Acb, AcbTm},
        reckless::Cif,
    };

    fn _marker_vec<Prog, C>()
    where
        for<'r> Prog: Eval<&'r Vec<C>, Vec<C>>,
    {
    }

    fn _marker_buff<Prog, C>()
    where
        for<'a> Prog: Eval<&'a mut Vec<C>, ()>,
    {
    }

    #[test]
    fn marker() {
        _marker_vec::<Program<Cif>, Cif>();
        _marker_buff::<RecklessProg, Cif>();
        _marker_vec::<RecklessProg, Cif>();
        _marker_vec::<RecklessProg, RecklessTm5>();
        _marker_vec::<Program<Cif>, RecklessTm5>();
        _marker_vec::<Program<Acb>, Acb>();
        _marker_vec::<AcbProg, Acb>();
        _marker_vec::<AcbProg, AcbTm<20>>();
        _marker_buff::<AcbProg, Acb>();
        _marker_buff::<AcbProg, AcbTm<20>>();
        _marker_vec::<Program<Acb>, AcbTm<20>>();
    }

    #[test]
    fn test1() {
        let ctx = circuit::Context::<i64>::default();
        let x = ctx.var("x");
        let y = ctx.var("y");
        let z = -(ctx.scalar(1) + &x + ctx.zero() - ctx.scalar(2) * &y).pow(25);

        let prog = Program::<i64>::compile(&[z.clone()][..], &[x.clone(), y.clone()][..]);
        let mut pt = std::collections::HashMap::default();
        pt.insert("x".to_string(), 1);
        pt.insert("y".to_string(), 2);

        let mut pt_circuit = std::collections::HashMap::default();
        pt_circuit.insert(x, 1);
        pt_circuit.insert(y, 2);

        assert!(prog.eval(&pt)[0] == z.eval(&pt_circuit).unwrap());
    }
}
