use crate::{
    flint::{Arb, Fmpq},
    reckless::{
        self,
        complex::arith::{_add, _mul_bounded, _mul_interval},
        with_rounding_up, Cif, Rif,
    },
    taylor_model::{
        tm::{self, Tm5},
        Tm,
    },
    traits::{
        interval::RealInterval,
        num::{MulWithRop, RefNeg},
        Interval, TaylorModel,
    },
};
use flint_bindings::arb_set_fmpq;
use nalgebra::Complex;
use num_traits::{One, Zero};
use std::ops::{Add, AddAssign, Mul, MulAssign, Neg};

#[derive(Debug, Clone, PartialEq)]
pub struct UnsafeRif(pub Rif);

#[derive(Debug, Clone, PartialEq, Copy)]
pub struct UnsafeCif(pub Cif);

impl Interval for UnsafeRif {
    fn diam(&self) -> f64 {
        self.0.diam()
    }

    fn mag(&self) -> f64 {
        self.0.mag()
    }

    fn mag_lower(&self) -> f64 {
        self.0.mag_lower()
    }

    fn mut_squash(&mut self) {
        self.0.mut_squash();
    }

    fn squash(&self) -> Self {
        Self(self.0.squash())
    }

    fn square(rad: f64) -> Self {
        Self(Rif::square(rad))
    }

    fn union(&self, other: &Self) -> Self {
        Self(self.0.union(&other.0))
    }
}

impl RealInterval for UnsafeRif {
    fn entire() -> Self {
        Self(Rif::entire())
    }

    fn from_inf_sup(inf: Self, sup: Self) -> Self {
        Self(Rif::from_inf_sup(inf.0, sup.0))
    }

    fn inf(&self) -> Self {
        Self(self.0.inf())
    }

    fn sup(&self) -> Self {
        Self(self.0.sup())
    }

    fn intersection(&self, other: &Self) -> Self {
        Self(self.0.intersection(other.0))
    }

    fn subset(&self, other: &Self) -> bool {
        self.0.subset(other.0)
    }
}

impl Add<&UnsafeCif> for UnsafeCif {
    type Output = Self;

    fn add(self, rhs: &UnsafeCif) -> Self::Output {
        Self(_add(self.0, rhs.0))
    }
}

impl<'l, 'r> Mul<&'r UnsafeCif> for &'l UnsafeCif {
    type Output = UnsafeCif;

    fn mul(self, rhs: &'r UnsafeCif) -> Self::Output {
        UnsafeCif(_mul_bounded(self.0, rhs.0))
    }
}

impl<'l, 'r> Add<&'r UnsafeCif> for &'l UnsafeCif {
    type Output = UnsafeCif;

    fn add(self, rhs: &'r UnsafeCif) -> Self::Output {
        UnsafeCif(_add(self.0, rhs.0))
    }
}

impl AddAssign for UnsafeCif {
    fn add_assign(&mut self, rhs: Self) {
        self.0 = _add(self.0, rhs.0);
    }
}

impl Mul<&UnsafeRif> for UnsafeCif {
    type Output = Self;

    fn mul(self, rhs: &UnsafeRif) -> Self::Output {
        Self(_mul_interval(self.0, rhs.0))
    }
}

impl MulAssign<&UnsafeRif> for UnsafeCif {
    fn mul_assign(&mut self, rhs: &UnsafeRif) {
        self.0 = _mul_interval(self.0, rhs.0);
    }
}

impl Mul for UnsafeCif {
    type Output = Self;

    fn mul(self, rhs: Self) -> Self::Output {
        Self(_mul_bounded(self.0, rhs.0))
    }
}

impl One for UnsafeCif {
    fn one() -> Self {
        Self(Cif::one())
    }

    fn is_one(&self) -> bool
    where
        Self: PartialEq,
    {
        self.0.is_one()
    }

    fn set_one(&mut self) {
        self.0.set_one();
    }
}

impl Zero for UnsafeCif {
    fn zero() -> Self {
        Self(Cif::zero())
    }

    fn is_zero(&self) -> bool {
        self.0.is_zero()
    }

    fn set_zero(&mut self) {
        self.0.set_zero();
    }
}

impl Add for UnsafeCif {
    type Output = Self;

    fn add(self, rhs: Self) -> Self::Output {
        Self(_add(self.0, rhs.0))
    }
}

impl AddAssign<&UnsafeCif> for UnsafeCif {
    fn add_assign(&mut self, rhs: &UnsafeCif) {
        self.0 = _add(self.0, rhs.0);
    }
}

impl MulAssign<&UnsafeCif> for UnsafeCif {
    fn mul_assign(&mut self, rhs: &UnsafeCif) {
        self.0 = _mul_bounded(self.0, rhs.0);
    }
}

impl RefNeg for UnsafeCif {
    fn ref_neg(&mut self) {
        self.0.ref_neg();
    }
}

impl From<&Complex<Fmpq>> for UnsafeCif {
    fn from(value: &Complex<Fmpq>) -> Self {
        Self(value.into())
    }
}

impl Neg for UnsafeCif {
    type Output = Self;

    fn neg(self) -> Self::Output {
        Self(-self.0)
    }
}

impl Interval for UnsafeCif {
    fn diam(&self) -> f64 {
        self.0.diam()
    }

    fn mag(&self) -> f64 {
        self.0.mag()
    }

    fn mag_lower(&self) -> f64 {
        self.0.mag_lower()
    }

    fn mut_squash(&mut self) {
        self.0.mut_squash();
    }

    fn squash(&self) -> Self {
        Self(self.0.squash())
    }

    fn square(rad: f64) -> Self {
        Self(Cif::square(rad))
    }

    fn union(&self, other: &Self) -> Self {
        Self(self.0.union(&other.0))
    }
}

impl MulWithRop<UnsafeCif> for UnsafeCif {
    type Output = Self;

    fn mul_with_rop(&self, rhs: &UnsafeCif, res: &mut Self::Output) {
        res.0 = _mul_bounded(self.0, rhs.0);
    }
}

#[derive(Debug, Clone, PartialEq)]
pub struct RecklessTm5(pub Tm5<Rif, Cif>);

impl TaylorModel for RecklessTm5 {
    type Coeffs = Cif;
    type Domain = Rif;

    fn order(&self) -> usize {
        5
    }

    fn from_vec(dom: &Self::Domain, coeffs: &Vec<Self::Coeffs>) -> Self {
        Self(Tm5::from_vec(dom, coeffs))
    }

    fn domain(&self) -> Self::Domain {
        self.0.domain()
    }

    fn set_domain(&mut self, dom: &Self::Domain) {
        self.0.set_domain(dom);
    }

    fn set_coeff(&mut self, coeff: Self::Coeffs, i: usize) {
        self.0.set_coeff(coeff, i);
    }

    fn get_coeff_mut(&mut self, i: usize) -> &mut Self::Coeffs {
        self.0.get_coeff_mut(i)
    }

    fn affine(a: &Self::Coeffs, b: &Self::Coeffs, domain: &Self::Domain) -> Self {
        Self(Tm5::affine(a, b, domain))
    }

    fn identity(domain: &Self::Domain) -> Self {
        Self(Tm5::identity(domain))
    }

    fn dominter(&mut self, dom: &Self::Domain) {
        self.0.dominter(dom);
    }

    fn enclosure(&self, dom: &Self::Domain) -> Self::Coeffs {
        with_rounding_up(|| self.0.enclosure(dom))
    }
}

impl AddAssign<&RecklessTm5> for RecklessTm5 {
    fn add_assign(&mut self, rhs: &RecklessTm5) {
        with_rounding_up(|| unsafe {
            *(self as *mut RecklessTm5 as *mut UnsafeRecklessTm5) += &*(rhs as *const RecklessTm5 as *const UnsafeRecklessTm5);
        })
    }
}

impl MulAssign<&RecklessTm5> for RecklessTm5 {
    fn mul_assign(&mut self, rhs: &RecklessTm5) {
        with_rounding_up(|| unsafe {
            *(self as *mut RecklessTm5 as *mut UnsafeRecklessTm5) *= &*(rhs as *const RecklessTm5 as *const UnsafeRecklessTm5);
        })
    }
}

impl AddAssign<&reckless::Cif> for RecklessTm5 {
    fn add_assign(&mut self, rhs: &reckless::Cif) {
        with_rounding_up(|| unsafe {
            *(self as *mut RecklessTm5 as *mut UnsafeRecklessTm5) += &*(rhs as *const Cif as *const UnsafeCif)
        });
    }
}

impl MulAssign<&reckless::Cif> for RecklessTm5 {
    fn mul_assign(&mut self, rhs: &reckless::Cif) {
        with_rounding_up(|| unsafe {
            *(self as *mut RecklessTm5 as *mut UnsafeRecklessTm5) *= &*(rhs as *const Cif as *const UnsafeCif)
        });
    }
}

impl RefNeg for RecklessTm5 {
    fn ref_neg(&mut self) {
        self.0.ref_neg();
    }
}

impl Add for RecklessTm5 {
    type Output = Self;

    fn add(mut self, rhs: Self) -> Self::Output {
        self += &rhs;
        self
    }
}

impl Zero for RecklessTm5 {
    fn zero() -> Self {
        Self(Tm5::zero())
    }

    fn is_zero(&self) -> bool {
        self.0.is_zero()
    }

    fn set_zero(&mut self) {
        self.0.set_zero();
    }
}

impl From<reckless::Cif> for RecklessTm5 {
    fn from(value: reckless::Cif) -> Self {
        Self(value.into())
    }
}

impl MulWithRop<Cif> for RecklessTm5 {
    type Output = Self;

    fn mul_with_rop(&self, rhs: &Cif, res: &mut Self::Output) {
        *res = self.clone();
        *res *= rhs;
    }
}

impl AddAssign for RecklessTm5 {
    fn add_assign(&mut self, rhs: Self) {
        *self += &rhs;
    }
}

// pub type UnsafeRecklessTm5 = Tm5<UnsafeRif, UnsafeCif>;

#[cfg(test)]
mod unsafe_reckless_tm5_marker {
    use std::ops::{AddAssign, MulAssign};

    use super::UnsafeRecklessTm5;

    fn _marker_add_assign<T>()
    where
        for<'r> T: AddAssign<&'r T>,
    {
    }

    fn _marker_mul_assign<T>()
    where
        for<'r> T: MulAssign<&'r T>,
    {
    }

    #[test]
    fn test() {
        _marker_add_assign::<UnsafeRecklessTm5>();
        _marker_mul_assign::<UnsafeRecklessTm5>();
    }
}

#[derive(Clone, Debug, PartialEq)]
pub struct UnsafeRecklessTm5(pub Tm5<UnsafeRif, UnsafeCif>);

impl TaylorModel for UnsafeRecklessTm5 {
    type Domain = UnsafeRif;
    type Coeffs = UnsafeCif;

    fn affine(a: &Self::Coeffs, b: &Self::Coeffs, domain: &Self::Domain) -> Self {
        Self(Tm5::affine(a, b, domain))
    }

    fn domain(&self) -> Self::Domain {
        self.0.domain()
    }

    fn dominter(&mut self, dom: &Self::Domain) {
        self.0.dominter(dom);
    }

    fn enclosure(&self, dom: &Self::Domain) -> Self::Coeffs {
        self.0.enclosure(dom)
    }

    fn from_vec(dom: &Self::Domain, coeffs: &Vec<Self::Coeffs>) -> Self {
        Self(Tm5::from_vec(dom, coeffs))
    }

    fn order(&self) -> usize {
        5
    }

    fn get_coeff_mut(&mut self, i: usize) -> &mut Self::Coeffs {
        self.0.get_coeff_mut(i)
    }

    fn identity(domain: &Self::Domain) -> Self {
        Self(Tm5::identity(domain))
    }

    fn set_coeff(&mut self, coeff: Self::Coeffs, i: usize) {
        self.0.set_coeff(coeff, i);
    }

    fn set_domain(&mut self, dom: &Self::Domain) {
        self.0.set_domain(dom);
    }
}

impl AddAssign<&UnsafeRecklessTm5> for UnsafeRecklessTm5 {
    fn add_assign(&mut self, rhs: &UnsafeRecklessTm5) {
        self.0 += &rhs.0;

        for i in 0..5 {
            if !self.0 .0.coeffs[i].0.is_common() {
                self.0 .0.coeffs[4].0 = Cif::new(Rif::ENTIRE, Rif::ENTIRE);
            }
        }
    }
}

impl MulAssign<&UnsafeRecklessTm5> for UnsafeRecklessTm5 {
    fn mul_assign(&mut self, rhs: &UnsafeRecklessTm5) {
        let a4 = self.0 .0.coeffs[4].0.clone();
        self.0 *= &rhs.0;

        if !(a4.is_common()
            & rhs.0 .0.coeffs[4].0.is_common()
            & self.0 .0.coeffs[0].0.is_common()
            & self.0 .0.coeffs[1].0.is_common()
            & self.0 .0.coeffs[2].0.is_common()
            & self.0 .0.coeffs[3].0.is_common()
            & self.0 .0.coeffs[4].0.is_common())
        {
            self.0 .0.coeffs[4].0 = reckless::Cif::new(reckless::Rif::ENTIRE, reckless::Rif::ENTIRE);
        }
    }
}

impl AddAssign<&UnsafeCif> for UnsafeRecklessTm5 {
    fn add_assign(&mut self, rhs: &UnsafeCif) {
        self.0 += rhs;
    }
}

impl MulAssign<&UnsafeCif> for UnsafeRecklessTm5 {
    fn mul_assign(&mut self, rhs: &UnsafeCif) {
        self.0 *= rhs;
    }
}

impl From<UnsafeCif> for UnsafeRecklessTm5 {
    fn from(value: UnsafeCif) -> Self {
        Self(value.into())
    }
}

impl Zero for UnsafeRecklessTm5 {
    fn zero() -> Self {
        Self(Tm5::zero())
    }

    fn is_zero(&self) -> bool {
        self.0.is_zero()
    }

    fn set_zero(&mut self) {
        self.0.set_zero();
    }
}

impl Add for UnsafeRecklessTm5 {
    type Output = Self;

    fn add(mut self, rhs: Self) -> Self::Output {
        self += &rhs;
        self
    }
}

impl RefNeg for UnsafeRecklessTm5 {
    fn ref_neg(&mut self) {
        self.0.ref_neg();
    }
}

impl Neg for UnsafeRecklessTm5 {
    type Output = Self;

    fn neg(mut self) -> Self::Output {
        self.0.ref_neg();
        self
    }
}

impl Mul for UnsafeRecklessTm5 {
    type Output = Self;

    fn mul(mut self, rhs: Self) -> Self::Output {
        self *= &rhs;
        self
    }
}

impl One for UnsafeRecklessTm5 {
    fn one() -> Self {
        Self(Tm5(Tm::one()))
    }

    fn is_one(&self) -> bool
    where
        Self: PartialEq,
    {
        self.0 .0.is_one()
    }

    fn set_one(&mut self) {
        self.0 .0.set_one();
    }
}

impl Add<Cif> for &RecklessTm5 {
    type Output = RecklessTm5;

    fn add(self, rhs: Cif) -> Self::Output {
        let mut res = self.clone();
        res += &rhs;
        res
    }
}

impl MulWithRop<UnsafeCif> for UnsafeRecklessTm5 {
    type Output = Self;

    fn mul_with_rop(&self, rhs: &UnsafeCif, res: &mut Self::Output) {
        *res = self.clone();
        *res *= rhs;
    }
}

impl From<&Fmpq> for Rif {
    fn from(value: &Fmpq) -> Self {
        let mut res = Arb::default();
        unsafe {
            arb_set_fmpq(&mut res.0, &value.0, 106);
        }
        res.get_covering_reckless_interval()
    }
}

#[cfg(test)]
mod tests {
    use crate::{flint::Fmpq, reckless::Rif};

    #[test]
    fn from_fmpq() {
        let x = Fmpq::from_machine_int(1, 3);
        let box_x: Rif = (&x).into();
        dbg!(box_x);
    }
}
