use super::{
    distort::{ArbDistort, Distort, RecklessDistort},
    predictor::{ArbPredictor, Predictor, RecklessPredictor, RecklessPredictor5},
};
use crate::{
    box_mat::BoxMat,
    box_vec::BoxVec,
    flint::{Acb, AcbMat, AcbVec, AcbVecTm, Arb, Fmpq},
    precision_ctx::{AdaptivePrecCtx, FixedPrecCtx, PrecCtx},
    prog::{AcbProg, Eval, ProgramTrait, RecklessProg},
    projective::{FwBuf, MooreBox, ParametricSystem, ProjectivePoint},
    reckless::{Cif, Rif},
    reckless_extra::RecklessTm5,
    taylor_model::{tm::RecklessTm, Tm},
    traits::{interval::RealInterval, matrix::Matrix, num::MulWithRop, vector::Vector, Interval, TaylorModel},
};
use anyhow::bail;
use itertools::izip;
use libm::frexp;
use nalgebra::Complex;
use num_traits::{Inv, One, Pow, Zero};
use std::{
    fmt::Debug,
    marker::PhantomData,
    ops::{Add, AddAssign, Div, Mul, Neg, Sub},
};

pub trait Extend {
    type Sys;
    type C;
    type VecC;
    type MatC;
    type Tm;
    type VecTm;

    fn extend(
        sys: &Self::Sys,
        mbox: &mut MooreBox<Self::VecC, Self::MatC>,
        step_size: f64,
        prev_pt: &mut Option<Self::VecC>,
        prev_speed: &mut Option<Self::VecC>,
    ) -> anyhow::Result<(f64, ProjectivePoint<Self::VecTm>)>;
}

pub struct GExtend<Prog, Ctx, R, C, VecC, MatC, Tm, VecTm, MatTm, GPredictor, GDistort> {
    phantom: PhantomData<(Prog, Ctx, R, C, VecC, MatC, Tm, VecTm, MatTm, GPredictor, GDistort)>,
}

impl<Prog, Ctx, R, C, VecC, MatC, Tm, VecTm, MatTm, GPredictor, GDistort> Extend
    for GExtend<Prog, Ctx, R, C, VecC, MatC, Tm, VecTm, MatTm, GPredictor, GDistort>
where
    for<'r1, 'r2> Prog: Eval<(&'r1 VecC, &'r2 VecC), Vec<C>> + ProgramTrait<Coeffs = C>,
    ParametricSystem<Prog>: FwBuf<VecTm, S = Tm> + FwBuf<VecC, S = C>,
    for<'r> R: Debug
        + From<f64>
        + Interval
        + Mul<R, Output = R>
        + RealInterval
        + Clone
        + Inv<Output = R>
        + Zero
        + Debug
        + Sub<&'r R, Output = R>,
    for<'l, 'r> &'l R: Add<R, Output = R> + Pow<usize, Output = R> + Sub<&'r R, Output = R>,
    for<'r> C: Clone + Zero + One + From<R> + MulWithRop<C, Output = C> + Interval + Debug + From<&'r Complex<Fmpq>>,
    for<'l> &'l C: TryInto<R, Error = anyhow::Error>,
    for<'r> VecC: Vector<T = C>
        + Interval
        + Neg<Output = VecC>
        + Div<C, Output = VecC>
        + Debug
        + Add<&'r VecC, Output = VecC>
        + Mul<&'r C, Output = VecC>,
    for<'l> &'l VecC: Sub<VecC, Output = VecC>,
    MatC: Interval + Matrix<Coeffs = C>,
    for<'l, 'r> &'l MatC: Mul<&'r VecC, Output = VecC>,
    for<'r> Tm: TaylorModel<Domain = R, Coeffs = C> + MulWithRop<C, Output = Tm> + Zero + Debug + AddAssign<&'r Tm>,
    for<'r> VecTm: TaylorModel<Domain = R, Coeffs = VecC> + Clone + Vector<T = Tm> + Debug + AddAssign<&'r VecTm>,
    Ctx: PrecCtx,
    GPredictor: Predictor<R = R, VecC = VecC, VecTm = VecTm>,
    GDistort: Distort<VecTm, MatTm, Sys = ParametricSystem<Prog>, S = Tm, MatC = MatC>
        + Distort<VecC, MatC, Sys = ParametricSystem<Prog>, S = C, MatC = MatC>,
{
    type Sys = ParametricSystem<Prog>;
    type C = C;
    type VecC = VecC;
    type MatC = MatC;
    type Tm = Tm;
    type VecTm = VecTm;

    fn extend(
        sys: &Self::Sys,
        mbox: &mut MooreBox<Self::VecC, Self::MatC>,
        mut step_size: f64,
        prev_pt: &mut Option<VecC>,
        prev_speed: &mut Option<VecC>,
    ) -> anyhow::Result<(f64, ProjectivePoint<Self::VecTm>)> {
        // Checking input in debug mode
        debug_assert!(mbox.params.len() == 1, "Moore box params must have len 1");
        debug_assert!(
            (&mbox.mat * &(sys.f(&mbox.params, &mbox.pt)) * &<f64 as Into<R>>::into(mbox.rad).inv().into()
                + &GDistort::distort(sys, &mbox.params, &mbox.pt, &mbox.mat, mbox.rad,))
                .mag()
                <= 0.875,
            "Mbox must satisfy Krawczyk's criterion in the selected arithmetic"
        );

        mbox.mut_squash(); // Sanity squash (that does not change the property of mbox)
        let t = (&mbox.params[0]).try_into().unwrap(); // t is the time at wich the moore box is valid (tight)

        // Predictor construction:
        //TODO: review system
        let fdot: VecC = sys.fdot[mbox.pt.chart].eval((&mbox.params, &mbox.pt.coord)).into();
        let mut speed = -(&mbox.mat * &fdot);
        speed.mut_squash();
        let save_pt = mbox.pt.coord.clone();
        let mut point_tm = ProjectivePoint {
            coord: GPredictor::predictor(&mbox.pt.coord, &speed, prev_pt, prev_speed, step_size, R::entire()),
            chart: mbox.pt.chart,
        };

        step_size *= 1.25;

        let ratios = [1.0, 0.875, 0.75, 0.625];
        loop {
            // Target times we wish to test. We also build intervals for each target time t tightly containing [0, t]
            let step_sizes = ratios.map(|ratio| step_size * ratio); // [s₁, ..., s₄] ≈ [s, ..., 0.625⋅s] step_sizes
            let target_ts = step_sizes.map(|step_size| (t.clone() + step_size.into()).sup()); // [t₁, ..., t₄] ≈ [t + s₁, ..., t + s₄]
            let domains = target_ts.clone().map(|target_t| R::from_inf_sup(0.0.into(), target_t - &t)); // domains containing tᵢ - t ≈ sᵢ
            let mut domain = domains.iter().fold(R::zero(), |accu, domain| accu.union(domain)); // ⋃ domains
            point_tm.coord.set_domain(&domain);
            let param = VecTm::affine(&mbox.params, &VecC::from_value(mbox.params.len(), C::one()), &domain); // t + η

            // Compute Krawczyk's operator with Tms
            let f = sys.f(&param, &point_tm);
            let mut krawczyk: Self::VecTm = VecTm::zeros(sys.dim);
            let mut auxi = C::zero();
            let mut auxi_tm = Tm::zero();
            for i in 0..sys.dim {
                for j in 0..sys.dim {
                    mbox.mat[(i, j)].mul_with_rop(&C::from(R::from(mbox.rad).inv()), &mut auxi);
                    f[j].mul_with_rop(&auxi, &mut auxi_tm);
                    krawczyk[i] += &auxi_tm;
                }
                domain = domain.intersection(&krawczyk[i].domain());
            }
            krawczyk += &GDistort::distort(sys, &param, &point_tm, &mbox.mat, mbox.rad);

            // Extension certification
            for (domain, target_t, step_size) in izip!(&domains, &target_ts, step_sizes) {
                if krawczyk.enclosure(domain).mag() <= 0.875 {
                    let target_pt = point_tm.coord.enclosure(&(target_t - &t));
                    while target_pt.diam() * 8.0 >= mbox.rad {
                        if Ctx::set_prec(Ctx::get_prec() + 1).is_err() {
                            bail!("innacurate taylor model evaluation at step size")
                        }
                    }
                    point_tm.coord.set_domain(domain);
                    *prev_pt = Some(save_pt);
                    *prev_speed = Some(speed);
                    mbox.params[0] = target_t.clone().into();
                    mbox.pt = ProjectivePoint {
                        coord: target_pt,
                        chart: mbox.pt.chart,
                    };
                    return Ok((step_size, point_tm));
                }
            }

            step_size *= 0.3;
            if -frexp(step_size).1 as i64 >= 1000 {
                bail!("getting close to the smallest representable f64");
            }
            if -frexp(step_size).1 as i64 >= Ctx::get_prec() && Ctx::set_prec(Ctx::get_prec() + 1).is_err() {
                bail!("step size is {} which looks ridiculously small", step_size);
            }
        }
    }
}

pub type RecklessExtend<const N: usize> = GExtend<
    RecklessProg,
    FixedPrecCtx,
    Rif,
    Cif,
    BoxVec<Cif>,
    BoxMat<Cif>,
    RecklessTm<N>,
    BoxVec<RecklessTm<N>>,
    BoxMat<RecklessTm<N>>,
    RecklessPredictor<N>,
    RecklessDistort,
>;

pub type RecklessExtend5 = GExtend<
    RecklessProg,
    FixedPrecCtx,
    Rif,
    Cif,
    BoxVec<Cif>,
    BoxMat<Cif>,
    RecklessTm5,
    BoxVec<RecklessTm5>,
    BoxMat<RecklessTm5>,
    RecklessPredictor5,
    RecklessDistort,
>;

pub type ArbExtend<const N: usize> = GExtend<
    AcbProg,
    AdaptivePrecCtx,
    Arb,
    Acb,
    AcbVec,
    AcbMat,
    Tm<N, Arb, Acb>,
    AcbVecTm<N>,
    BoxMat<Tm<N, Arb, Acb>>,
    ArbPredictor<N>,
    ArbDistort,
>;

#[cfg(test)]
mod marker_tests {
    use super::{ArbExtend, Extend, RecklessExtend, RecklessExtend5};

    fn _marker<T: Extend>() {}

    #[test]
    fn marker() {
        _marker::<RecklessExtend<20>>();
        _marker::<RecklessExtend5>();
        _marker::<ArbExtend<10>>();
    }
}
