use super::distort::{ArbDistort, Distort, RecklessDistort};
use crate::{
    box_mat::BoxMat,
    box_vec::BoxVec,
    flint::{Acb, AcbMat, AcbVec, Arb},
    precision_ctx::set_prec,
    prog::{AcbProg, RecklessProg},
    projective::{FwBuf, InvDf, MooreBox, ParametricSystem, ProjectivePoint},
    reckless::{Cif, Rif},
    traits::{system::AcbRecklessSys, Interval},
};
use anyhow::bail;
use num_traits::Inv;
use std::{
    fmt::Debug,
    marker::PhantomData,
    ops::{Add, Mul},
};

pub trait FindMooreBox {
    type Sys;
    type C;
    type VecC;
    type MatC;

    fn find_moore_box(
        sys: &Self::Sys,
        params: &Self::VecC,
        pt: ProjectivePoint<Self::VecC>,
    ) -> anyhow::Result<MooreBox<Self::VecC, Self::MatC>>;
}

pub struct GFindMooreBox<Prog, R, C, VecC, MatC, GDistort> {
    phantom: PhantomData<(Prog, R, C, VecC, MatC, GDistort)>,
}

impl<Prog, R, C, VecC, MatC, GDistort> FindMooreBox for GFindMooreBox<Prog, R, C, VecC, MatC, GDistort>
where
    R: From<f64> + Inv<Output = R>,
    C: From<R> + Debug,
    for<'r> VecC: Interval + Mul<&'r C, Output = VecC> + Add<VecC, Output = VecC> + Add<&'r VecC, Output = VecC>,
    GDistort: Distort<VecC, MatC, Sys = ParametricSystem<Prog>, MatC = MatC, S = C>,
    ParametricSystem<Prog>: InvDf<VecC, MatC, S = C> + FwBuf<VecC, S = C>,
    for<'l> &'l MatC: Mul<VecC, Output = VecC>,
{
    type Sys = ParametricSystem<Prog>;
    type C = C;
    type VecC = VecC;
    type MatC = MatC;

    fn find_moore_box(
        sys: &Self::Sys,
        params: &Self::VecC,
        pt: ProjectivePoint<Self::VecC>,
    ) -> anyhow::Result<MooreBox<Self::VecC, Self::MatC>> {
        let mut rad = 1.0;
        // inv_df: ≈ dfₜ(x)⁻¹
        let inv_df = sys.inv_df(params, &pt)?;
        for _ in 1..100 {
            // distort: [A⋅dfₜ(x + rB) - 1]⋅B
            let distort = GDistort::distort(sys, params, &pt, &inv_df, rad);
            // dbg!(distort.mag());
            if distort.mag() <= 0.5 {
                // condition: ||A⋅fₜ(x)/r + [A⋅dfₜ(x + rB) - 1]⋅B|| ≤ 0.875
                return Ok(MooreBox {
                    params: params.clone(),
                    pt,
                    mat: inv_df,
                    rad,
                });
                if ((&inv_df * sys.f(params, &pt)) * &C::from(R::from(rad).inv()) + distort).mag() <= 0.875 {
                    return Ok(MooreBox {
                        params: params.clone(),
                        pt,
                        mat: inv_df,
                        rad,
                    });
                } else {
                    bail!("could not find a Moore box around this approximate zero");
                }
            }
            rad *= 0.5;
        }
        return Ok(MooreBox {
            params: params.clone(),
            pt,
            mat: inv_df,
            rad,
        });
        bail!("the Moore box construction procedure has still not finished after 100 iterations, usually means that roundoff errors dominate");
    }
}

pub type RecklessFindMooreBox = GFindMooreBox<RecklessProg, Rif, Cif, BoxVec<Cif>, BoxMat<Cif>, RecklessDistort>;

pub type ArbFindMooreBox = GFindMooreBox<AcbProg, Arb, Acb, AcbVec, AcbMat, ArbDistort>;

pub struct ArbRecklessFindMooreBox {}

impl FindMooreBox for ArbRecklessFindMooreBox {
    type Sys = AcbRecklessSys;
    type C = Acb;
    type VecC = AcbVec;
    type MatC = AcbMat;

    fn find_moore_box(
        sys: &Self::Sys,
        params: &Self::VecC,
        pt: ProjectivePoint<Self::VecC>,
    ) -> anyhow::Result<MooreBox<Self::VecC, Self::MatC>> {
        ArbFindMooreBox::find_moore_box(&sys.acb_sys, params, pt)
    }
}

#[cfg(test)]
mod marker_tests {

    use super::{ArbFindMooreBox, ArbRecklessFindMooreBox, FindMooreBox, RecklessFindMooreBox};

    fn _marker<T: FindMooreBox>() {}

    #[test]
    fn marker() {
        _marker::<RecklessFindMooreBox>();
        _marker::<ArbFindMooreBox>();
        _marker::<ArbRecklessFindMooreBox>();
    }
}
