use anyhow::bail;
use num_traits::{One, Zero};

use super::{
    extend::{ArbExtend, Extend, RecklessExtend, RecklessExtend5},
    refine::{ArbRefine, RecklessRefine, Refine},
    to_chart::{ArbToChart, RecklessToChart, ToChart},
};
use crate::{
    box_mat::BoxMat,
    box_vec::BoxVec,
    flint::{Acb, AcbMat, AcbTm, AcbVec, AcbVecTm},
    foa::{FoA, FoATrait},
    precision_ctx::{get_prec, AdaptivePrecCtx, FixedPrecCtx, PrecCtx},
    prog::{AcbProg, RecklessProg},
    projective::{MooreBox, ParametricSystem, ProjectivePoint},
    reckless::Cif,
    reckless_extra::RecklessTm5,
    taylor_model::tm::RecklessTm,
    traits::{system::AcbRecklessSys, vector::Vector, Interval},
};
use std::{fmt::Debug, marker::PhantomData};

pub trait OneStep {
    type Sys;
    type C;
    type VecC;
    type MatC;
    type Tm;
    type VecTm;

    fn one_step(
        sys: &Self::Sys,
        mbox: &mut MooreBox<Self::VecC, Self::MatC>,
        step_size: f64,
        previous_pt: &mut Option<Self::VecC>,
        previous_speed: &mut Option<Self::VecC>,
    ) -> anyhow::Result<(f64, ProjectivePoint<Self::VecTm>)>;
}

pub struct GOneStep<GRefine, GExtend, GToChart, Ctx, Prog, C, VecC, MatC, Tm, VecTm> {
    phantom: PhantomData<(GRefine, GExtend, GToChart, Ctx, Prog, C, VecC, MatC, Tm, VecTm)>,
}

impl<GRefine, GExtend, GToChart, Ctx, Prog, C, VecC, MatC, Tm, VecTm> OneStep
    for GOneStep<GRefine, GExtend, GToChart, Ctx, Prog, C, VecC, MatC, Tm, VecTm>
where
    Ctx: PrecCtx,
    GRefine: Refine<Sys = ParametricSystem<Prog>, C = C, VecC = VecC, MatC = MatC>,
    GExtend: Extend<Sys = ParametricSystem<Prog>, C = C, VecC = VecC, MatC = MatC, Tm = Tm, VecTm = VecTm>,
    GToChart: ToChart<MatC = MatC, VecC = VecC, Sys = ParametricSystem<Prog>, C = C>,
    VecC: Interval + Debug + Vector,
    VecC::T: Clone + One + Interval + Zero,
    MatC: Interval,
{
    type Sys = ParametricSystem<Prog>;
    type C = C;
    type VecC = VecC;
    type MatC = MatC;
    type Tm = Tm;
    type VecTm = VecTm;

    fn one_step(
        sys: &Self::Sys,
        mbox: &mut MooreBox<Self::VecC, Self::MatC>,
        step_size: f64,
        previous_pt: &mut Option<VecC>,
        previous_speed: &mut Option<VecC>,
    ) -> anyhow::Result<(f64, ProjectivePoint<Self::VecTm>)> {
        if mbox.pt.coord.mag() > sys.chart_switch_threshold {
            #[cfg(debug_assertions)]
            println!("Changing chart ({})...", mbox.pt.chart);
            GToChart::to_chart(sys, mbox, mbox.pt.canonical_chart(), 0.75)?;
            mbox.mut_squash();
            *previous_pt = None;
            *previous_speed = None;
        }
        GRefine::refine(sys, mbox, 0.25)?;
        mbox.mut_squash();
        GExtend::extend(sys, mbox, step_size, previous_pt, previous_speed)
    }
}

pub type RecklessOneStep5 = GOneStep<
    RecklessRefine,
    RecklessExtend5,
    RecklessToChart,
    FixedPrecCtx,
    RecklessProg,
    Cif,
    BoxVec<Cif>,
    BoxMat<Cif>,
    RecklessTm5,
    BoxVec<RecklessTm5>,
>;

pub type RecklessOneStep<const N: usize> = GOneStep<
    RecklessRefine,
    RecklessExtend<N>,
    RecklessToChart,
    FixedPrecCtx,
    RecklessProg,
    Cif,
    BoxVec<Cif>,
    BoxMat<Cif>,
    RecklessTm<N>,
    BoxVec<RecklessTm<N>>,
>;

pub type ArbOneStep<const N: usize> =
    GOneStep<ArbRefine, ArbExtend<N>, ArbToChart, AdaptivePrecCtx, AcbProg, Acb, AcbVec, AcbMat, AcbTm<N>, AcbVecTm<N>>;

pub struct ArbRecklessOneStep {}

impl OneStep for ArbRecklessOneStep {
    type Sys = AcbRecklessSys;
    type C = Acb;
    type VecC = AcbVec;
    type MatC = AcbMat;
    type Tm = AcbTm<5>;
    type VecTm = AcbVecTm<5>;

    fn one_step(
        sys: &Self::Sys,
        mbox: &mut MooreBox<Self::VecC, Self::MatC>,
        step_size: f64,
        previous_pt: &mut Option<Self::VecC>,
        previous_speed: &mut Option<Self::VecC>,
    ) -> anyhow::Result<(f64, ProjectivePoint<Self::VecTm>)> {
        let current_prec = get_prec();
        if current_prec <= 53 {
            if let Ok(mut reckless_mbox) = (&*mbox).try_into() {
                // dbg!("Conversion sucessful. Going reckless !");
                let mut reckless_prev_pt: Option<BoxVec<Cif>> = match previous_pt {
                    None => None,
                    Some(v) => Some(
                        v.into_iter()
                            .map(|c| c.get_covering_reckless_interval())
                            .collect::<Vec<Cif>>()
                            .into(),
                    ),
                };
                let mut reckless_prev_speed: Option<BoxVec<Cif>> = match previous_speed {
                    None => None,
                    Some(v) => Some(
                        v.into_iter()
                            .map(|c| c.get_covering_reckless_interval())
                            .collect::<Vec<Cif>>()
                            .into(),
                    ),
                };

                match RecklessOneStep5::one_step(
                    &sys.reckless_sys,
                    &mut reckless_mbox,
                    step_size,
                    &mut reckless_prev_pt,
                    &mut reckless_prev_speed,
                ) {
                    Err(_) => return ArbOneStep::one_step(&sys.acb_sys, mbox, step_size, previous_pt, previous_speed),
                    Ok((s, _)) => {
                        *previous_pt = match reckless_prev_pt {
                            None => None,
                            Some(v) => Some(v.0.iter().map(|c| c.get_included_acb()).collect::<Vec<Acb>>().into()),
                        };
                        *previous_speed = match reckless_prev_speed {
                            None => None,
                            Some(v) => Some(v.0.iter().map(|c| c.get_included_acb()).collect::<Vec<Acb>>().into()),
                        };
                        *mbox = (&reckless_mbox).into();
                        return Ok((
                            s,
                            ProjectivePoint {
                                coord: AcbVecTm::zeros(1),
                                chart: 0,
                            },
                        ));
                    }
                }
            }
        }
        ArbOneStep::one_step(&sys.acb_sys, mbox, step_size, previous_pt, previous_speed)
    }
}

impl<F, A> OneStep for FoA<F, A>
where
    F: OneStep,
    A: OneStep,
    FoA<F::VecC, A::VecC>: FoATrait<F = F::VecC, A = A::VecC>,
{
    type Sys = FoA<F::Sys, A::Sys>;
    type C = FoA<F::C, A::C>;
    type VecC = FoA<F::VecC, A::VecC>;
    type MatC = FoA<F::MatC, A::MatC>;
    type Tm = FoA<A::Tm, A::Tm>;
    type VecTm = FoA<F::VecTm, A::VecTm>;

    fn one_step(
        sys: &Self::Sys,
        mbox: &mut MooreBox<Self::VecC, Self::MatC>,
        step_size: f64,
        previous_pt: &mut Option<Self::VecC>,
        previous_speed: &mut Option<Self::VecC>,
    ) -> anyhow::Result<(f64, ProjectivePoint<Self::VecTm>)> {
        let try_steps = || -> Result<(), anyhow::Error> {
            mbox.params.try_compute_fast()?;
            Ok(())
        };
        todo!()
    }
}

#[cfg(test)]
mod marker_tests {
    use super::{ArbOneStep, ArbRecklessOneStep, OneStep, RecklessOneStep, RecklessOneStep5};

    fn _marker<T: OneStep>() {}

    #[test]
    fn marker() {
        _marker::<RecklessOneStep5>();
        _marker::<RecklessOneStep<20>>();
        _marker::<ArbOneStep<20>>();
        _marker::<ArbRecklessOneStep>();
    }
}
