// Copyright 2024 Pierre Lairez
// SPDX-License-Identifier: GPL-3.0-or-later

use std::{path::PathBuf, time::Instant};

use algpath::{
    flint::Acb,
    traits::{
        system::AcbSys,
        track::{
            find_moore_box::{ArbFindMooreBox, FindMooreBox},
            ArbTrack, Track,
        },
    },
};

fn track_with_input(path: &str, homogenize: bool, horner: bool) {
    println!("Testing {} (homogenize = {}, horner = {})", path, homogenize, horner);
    let path = PathBuf::from(path);
    let mut conf = algpath::cli::Configuration {
        path: Some(path),
        ..Default::default()
    };

    let start = Instant::now();
    conf.homogenize = homogenize;
    conf.horner = horner;

    let mut input_problem = algpath::cli::Problem::new_from_input(&conf).unwrap();
    let exact_sys = input_problem.build_parametric_system(&conf).unwrap();
    let mut sys: AcbSys = exact_sys.into();

    let _path = &input_problem.path;
    let path: Vec<Vec<Acb>> = _path
        .clone()
        .into_iter()
        .map(|v| v.into_iter().map(|c| c.into()).collect())
        .collect();

    let _fiber = &input_problem.fiber;
    let fiber: Vec<Vec<Acb>> = _fiber
        .clone()
        .into_iter()
        .map(|v| v.into_iter().map(|c| c.into()).collect())
        .collect();

    for point in &fiber {
        let proj_point = algpath::projective::ProjectivePoint {
            coord: point.clone().into(),
            chart: 0,
        };
        let mut mbox = ArbFindMooreBox::find_moore_box(&mut sys, &path[0].clone().into(), proj_point).unwrap();

        let _tr = ArbTrack::<5>::track(&mut sys, &mut mbox).unwrap();
        println!("Number of steps: {}", _tr.len());
    }
    let duration = start.elapsed();
    println!("Time Elapsed: {:?}", duration);
}

#[test]
fn track() {
    for homogenize in [true, false] {
        for horner in [true, false] {
            track_with_input("data/2-2_1.json", homogenize, horner);
            track_with_input("data/prec_test.json", homogenize, horner);
        }
    }
    track_with_input("data/prec_test_2.json", false, false);
    track_with_input("data/prec_test_3.json", false, false);
    // track_with_input("data/3-3-3-3_1.json", false, false);
}
